import os
import numpy as np
import pytest
import xarray as xr
import pandas as pd
import cf_units as cf

import geokube.core.coord_system as crs
from geokube.backend import open_datacube, open_dataset
from geokube.core.axis import Axis, AxisType
from geokube.core.coord_system import GeogCS, RegularLatLon
from geokube.core.coordinate import Coordinate, CoordinateType
from geokube.core.datacube import DataCube
from geokube.core.domain import (
    Domain,
    DomainType,
    GeodeticGrid,
    GeodeticPoints,
)
from geokube.core.enums import LongitudeConvention, MethodType
from geokube.core.field import Field
from geokube.core.unit import Unit
from geokube.core.variable import Variable
from geokube.utils import util_methods
from tests import RES_PATH, clear_test_res, compare_dicts
from tests.fixtures import *


def test_from_xarray_with_point_domain(era5_point_domain):
    field = Field.from_xarray(era5_point_domain, ncvar="W_SO")
    assert "points" in field.domain["longitude"].dim_names
    assert "points" in field.domain["latitude"].dim_names
    assert "points" in field.domain[Axis("x")].dim_names
    assert "points" in field.domain[AxisType.Y].dim_names

    dset = field.to_xarray(encoding=False)
    assert "latitude" in dset.coords
    assert "lat" not in dset.coords
    assert "longitude" in dset.coords
    assert "lon" not in dset.coords
    assert "points" in dset.dims
    assert "points" in dset["grid_latitude"].dims
    assert "points" in dset["grid_longitude"].dims
    assert "points" in dset["latitude"].dims
    assert "points" in dset["longitude"].dims


def test_from_xarray_rotated_pole(era5_rotated_netcdf):
    field = Field.from_xarray(era5_rotated_netcdf, ncvar="TMIN_2M")

    assert field.name == "air_temperature"
    assert field.ncvar == "TMIN_2M"
    assert "height_2m" in field.domain
    assert "lon" in field.domain
    assert "lat" in field.domain
    assert "rlat" in field.domain
    assert "rlon" in field.domain
    assert field.domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )


def test_to_xarray_rotated_pole_without_encoding(era5_rotated_netcdf):
    field = Field.from_xarray(era5_rotated_netcdf, ncvar="TMIN_2M")
    xr_res = field.to_xarray(encoding=False)
    assert "air_temperature" in xr_res.data_vars
    assert "height" in xr_res.coords
    assert "time" in xr_res.coords
    assert "longitude" in xr_res.coords
    assert "latitude" in xr_res.coords
    assert "grid_latitude" in xr_res.coords
    assert "grid_longitude" in xr_res.coords
    assert "crs_rotated_latitude_longitude" in xr_res.coords
    assert "time" in xr_res["air_temperature"].dims
    assert "grid_latitude" in xr_res["air_temperature"].dims
    assert "grid_longitude" in xr_res["air_temperature"].dims
    assert "grid_mapping" in xr_res["air_temperature"].encoding
    assert xr_res["air_temperature"].encoding["grid_mapping"] == "crs_rotated_latitude_longitude"
    assert set(
        xr_res["air_temperature"].encoding["coordinates"].split(" ")
    ) == {
        "height",
        "latitude",
        "longitude",
    }
    assert "cell_methods" in xr_res["air_temperature"].attrs
    assert (
        xr_res["air_temperature"].attrs["cell_methods"]
        == era5_rotated_netcdf["TMIN_2M"].attrs["cell_methods"]
    )


def test_to_xarray_rotated_pole_with_encoding(era5_rotated_netcdf):
    field = Field.from_xarray(era5_rotated_netcdf, ncvar="TMIN_2M")
    xr_res = field.to_xarray(encoding=True)
    assert "TMIN_2M" in xr_res.data_vars
    assert "height_2m" in xr_res.coords
    assert "lon" in xr_res.coords
    assert "lat" in xr_res.coords
    assert "rlat" in xr_res.coords
    assert "rlon" in xr_res.coords
    assert "crs_rotated_latitude_longitude" in xr_res.coords
    assert "time" in xr_res["TMIN_2M"].dims
    assert "rlat" in xr_res["TMIN_2M"].dims
    assert "rlon" in xr_res["TMIN_2M"].dims
    assert "grid_mapping" in xr_res["TMIN_2M"].encoding
    assert xr_res["TMIN_2M"].encoding["grid_mapping"] == "crs_rotated_latitude_longitude"
    assert set(xr_res["TMIN_2M"].encoding["coordinates"].split(" ")) == {
        "height_2m",
        "lat",
        "lon",
    }
    assert "cell_methods" in xr_res["TMIN_2M"].attrs
    assert (
        xr_res["TMIN_2M"].attrs["cell_methods"]
        == era5_rotated_netcdf["TMIN_2M"].attrs["cell_methods"]
    )


def test_from_xarray_rotated_pole_with_mapping_and_id_pattern(
    era5_rotated_netcdf,
):
    field = Field.from_xarray(
        era5_rotated_netcdf,
        ncvar="TMIN_2M",
        id_pattern="prefix:{standard_name}",
        mapping={"rlat": {"name": "myrlat"}},
    )
    assert field.name == "prefix:air_temperature"
    assert field.ncvar == "TMIN_2M"
    assert "prefix:height" in field.domain._coords
    assert "prefix:longitude" in field.domain._coords
    assert "prefix:latitude" in field.domain._coords
    assert "myrlat" in field.domain._coords
    assert "prefix:grid_latitude" not in field.domain._coords
    assert "prefix:grid_longitude" in field.domain._coords
    assert field.domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )

    xr_res = field.to_xarray(encoding=False)
    assert "prefix:air_temperature" in xr_res.data_vars
    assert "myrlat" in xr_res.coords
    assert "prefix:grid_latitude" not in xr_res.coords
    assert "prefix:grid_longitude" in xr_res.coords
    assert "prefix:time" in xr_res.coords

    xr_res = field.to_xarray(encoding=True)
    assert "prefix:air_temperature" not in xr_res.data_vars
    assert "TMIN_2M" in xr_res.data_vars
    assert "myrlat" not in xr_res.coords
    assert "rlat" in xr_res.coords
    assert "prefix:grid_longitude" not in xr_res.coords
    assert "rlon" in xr_res.coords
    assert "prefix:time" not in xr_res.coords
    assert "time" in xr_res.coords


def test_from_xarray_curvilinear_grid(nemo_ocean_16):
    field = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    assert field.name == "vt"
    assert field.ncvar == "vt"
    assert field.units == Unit("degree_C m/s")
    assert str(field.cell_methods) == nemo_ocean_16["vt"].attrs["cell_methods"]
    assert "time" in field.domain._coords
    assert "depthv" in field.domain._coords
    assert "latitude" in field.domain._coords
    assert "longitude" in field.domain._coords
    assert "x" not in field.domain._coords
    assert "y" not in field.domain._coords
    assert isinstance(field.domain.crs, crs.CurvilinearGrid)

    assert field.domain["longitude"].dims[0].type == AxisType.Y
    assert field.domain["longitude"].dims[1].type == AxisType.X

    xr_res = field.to_xarray()
    assert xr_res["vt"].encoding["grid_mapping"] == "crs_curvilinear_grid"
    assert "crs_curvilinear_grid" in xr_res.coords


def test_from_xarray_regular_latlon(era5_netcdf):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    assert field._id_pattern is None
    assert field._mapping is None
    assert field.name == "tp"
    assert field.ncvar == "tp"
    assert field.units == Unit("m")


def test_from_xarray_regular_latlon_with_id_pattern(era5_netcdf):
    field = Field.from_xarray(
        era5_netcdf, ncvar="tp", id_pattern="{__ddsapi_name}"
    )
    assert field._id_pattern == "{__ddsapi_name}"
    assert field._mapping is None
    assert field.name == "total_precipitation"
    assert field.ncvar == "tp"
    assert field.units == Unit("m")


def test_from_xarray_regular_latlon_with_complex_id_pattern(era5_netcdf):
    field = Field.from_xarray(
        era5_netcdf, ncvar="tp", id_pattern="{units}__{long_name}"
    )
    assert field.name == "m__Total precipitation"
    assert field.ncvar == "tp"
    assert field.units == Unit("m")


def test_from_xarray_regular_latlon_with_mapping(era5_netcdf):
    field = Field.from_xarray(
        era5_netcdf, ncvar="tp", mapping={"tp": {"name": "tot_prep"}}
    )
    assert field._id_pattern is None
    assert field._mapping == {"tp": {"name": "tot_prep"}}
    assert field.name == "tot_prep"
    assert field.ncvar == "tp"
    assert field.units == Unit("m")


def test_geobbox_regular_latlon(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    with pytest.raises(ValueError):
        __ = tp.geobbox(
            north=10, south=-10, west=50, east=80, top=5, bottom=10
        )

    res = tp.geobbox(north=10, south=-10, west=50, east=80)
    assert np.all(res["latitude"].values <= 10)
    assert np.all(res["latitude"].values >= -10)
    assert np.all(res.latitude.values <= 10)
    assert np.all(res.latitude.values >= -10)

    assert np.all(res["longitude"].values <= 80)
    assert np.all(res["longitude"].values >= 50)
    assert np.all(res.longitude.values <= 80)
    assert np.all(res.longitude.values >= 50)

    dset = res.to_xarray(True)
    assert np.all(dset.latitude <= 10)
    assert np.all(dset.latitude >= -10)
    assert dset.latitude.attrs["units"] == "degrees_north"

    assert np.all(dset.longitude <= 80)
    assert np.all(dset.longitude >= 50)
    assert dset.longitude.attrs["units"] == "degrees_east"


def test_geobbox_regular_latlon_2(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    res = tp.geobbox(north=10, south=-10, west=-20, east=20)
    assert np.all(res["latitude"].values <= 10)
    assert np.all(res["latitude"].values >= -10)
    assert np.all(res.latitude.values <= 10)
    assert np.all(res.latitude.values >= -10)

    assert np.all(res["longitude"].values <= 20)
    assert np.all(res["longitude"].values >= -20)
    assert np.all(res.longitude.values <= 20)
    assert np.all(res.longitude.values >= -20)

    dset = res.to_xarray(True)
    assert np.all(dset.latitude <= 10)
    assert np.all(dset.latitude >= -10)
    assert dset.latitude.attrs["units"] == "degrees_north"

    assert np.all(dset.longitude <= 20)
    assert np.all(dset.longitude >= -20)
    assert dset.longitude.attrs["units"] == "degrees_east"


def test_geobbox_regular_latlon_3(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    res = tp.geobbox(north=10, west=2)
    assert np.all(res["latitude"].values <= 10)
    assert np.all(res.latitude.values <= 10)

    assert np.all(res["longitude"].values >= 2)
    assert np.all(res.longitude.values >= 2)

    dset = res.to_xarray(True)
    assert np.all(dset.latitude <= 10)
    assert dset.latitude.attrs["units"] == "degrees_north"

    assert np.all(dset.longitude >= 2)
    assert dset.longitude.attrs["units"] == "degrees_east"


def test_geobbox_regular_latlon_4(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    with pytest.raises(ValueError, match="'top' and 'bottom' must be None"):
        tp.geobbox(top=2)

    res = tp.geobbox(north=10)
    assert np.all(res["latitude"].values <= 10)
    assert np.all(res.latitude.values <= 10)

    dset = res.to_xarray(True)
    assert np.all(dset.latitude <= 10)
    assert dset.latitude.attrs["units"] == "degrees_north"


def test_locations_regular_latlon_1(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    lat, lon = 40, 15
    res = tp.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "latitude"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "longitude"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon)


def test_locations_regular_latlon_2(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    lat, lon = [35, 40], [15, 20]
    res = tp.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "latitude"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "longitude"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon)


def test_locations_regular_latlon_wrong_vertical(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")

    with pytest.raises(
        KeyError,
        match="Axis of type `AxisType.VERTICAL` does not exist in the domain!",
    ):
        tp.locations(latitude=[35, 40], longitude=[15, 20], vertical=[1, 2])


def test_geobbox_rotated_pole(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    assert wso.latitude.name == "latitude"
    assert wso.latitude.ncvar == "lat"
    assert wso.longitude.name == "longitude"
    assert wso.longitude.ncvar == "lon"

    res = wso.geobbox(north=40, south=38, west=16, east=19)
    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "lon"
    assert res.domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values >= 38) / W > 0.95
    assert np.sum(res.latitude.values <= 40) / W > 0.95
    assert np.sum(res.longitude.values >= 16) / W > 0.95
    assert np.sum(res.longitude.values <= 19) / W > 0.95

    dset = res.to_xarray(encoding=True)
    assert "W_SO" in dset.data_vars
    assert "rlat" in dset.coords
    assert "rlon" in dset.coords
    assert "lat" in dset
    assert dset.lat.attrs["units"] == "degrees_north"
    assert "lon" in dset
    assert dset.lon.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords

    dset = res.to_xarray(False)
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in dset.data_vars
    assert "latitude" in dset
    assert dset.latitude.attrs["units"] == "degrees_north"
    assert "longitude" in dset
    assert dset.longitude.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords


def test_geobbox_rotated_pole_partial_arguments_1(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")

    res = wso.geobbox(north=40, west=16)
    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "lon"
    assert res.domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values <= 40) / W > 0.95
    assert np.sum(res.longitude.values >= 16) / W > 0.95

    dset = res.to_xarray(encoding=True)
    assert "W_SO" in dset.data_vars
    assert "rlat" in dset.coords
    assert "rlon" in dset.coords
    assert "lat" in dset
    assert dset.lat.attrs["units"] == "degrees_north"
    assert "lon" in dset
    assert dset.lon.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords

    dset = res.to_xarray(False)
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in dset.data_vars
    assert "latitude" in dset
    assert dset.latitude.attrs["units"] == "degrees_north"
    assert "longitude" in dset
    assert dset.longitude.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords


def test_geobbox_rotated_pole_partial_arguments_2(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")

    res = wso.geobbox(north=40)
    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "lon"
    assert res.domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values <= 40) / W > 0.95

    dset = res.to_xarray(encoding=True)
    assert "W_SO" in dset.data_vars
    assert "rlat" in dset.coords
    assert "rlon" in dset.coords
    assert "lat" in dset
    assert dset.lat.attrs["units"] == "degrees_north"
    assert "lon" in dset
    assert dset.lon.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords

    dset = res.to_xarray(False)
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in dset.data_vars
    assert "latitude" in dset
    assert dset.latitude.attrs["units"] == "degrees_north"
    assert "longitude" in dset
    assert dset.longitude.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords


def test_locations_rotated_pole_1(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")

    lat, lon = 40, 15
    res = wso.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "lon"
    assert res.domain.type.name == "POINTS"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)


def test_locations_rotated_pole_2(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")

    lat, lon = [38, 40], [15, 18]
    res = wso.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "lon"
    assert res.domain.type.name == "POINTS"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)


def test_geobbox_curvilinear_grid_all(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    assert vt.latitude.name == "latitude"
    assert vt.longitude.name == "longitude"
    assert vt.latitude.ncvar == "nav_lat"
    assert vt.longitude.ncvar == "nav_lon"
    assert vt.vertical.name == "depthv"
    assert vt.vertical.ncvar == "depthv"

    res = vt.geobbox(
        north=-19, south=-22, west=-115, east=-110, bottom=-5, top=-1
    )
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values >= -22) / W > 0.95
    assert np.sum(res.latitude.values <= -19) / W > 0.95
    assert np.sum(res.longitude.values >= -115) / W > 0.95
    assert np.sum(res.longitude.values <= -110) / W > 0.95
    assert res.vertical.values.size == 5
    assert np.all((res.vertical.values >= 1) & (res.vertical.values <= 5))


def test_geobbox_curvilinear_grid_all_wrong_vertical(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(
        north=-19, south=-22, west=-115, east=-110, bottom=1, top=5
    )
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values >= -22) / W > 0.95
    assert np.sum(res.latitude.values <= -19) / W > 0.95
    assert np.sum(res.longitude.values >= -115) / W > 0.95
    assert np.sum(res.longitude.values <= -110) / W > 0.95
    assert res.vertical.values.size == 0


def test_geobbox_curvilinear_grid_horizontal_all(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(north=-19, south=-22, west=-115, east=-110)
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values >= -22) / W > 0.95
    assert np.sum(res.latitude.values <= -19) / W > 0.95
    assert np.sum(res.longitude.values >= -115) / W > 0.95
    assert np.sum(res.longitude.values <= -110) / W > 0.95


def test_geobbox_curvilinear_grid_horizontal_partial_1(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(north=-19, west=-115)
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values <= -19) / W > 0.95
    assert np.sum(res.longitude.values >= -115) / W > 0.95


def test_geobbox_curvilinear_grid_horizontal_partial_2(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(north=-19)
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    W = np.prod(res.latitude.shape)
    assert np.sum(res.latitude.values <= -19) / W > 0.95


def test_geobbox_curvilinear_grid_vertical_both(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(bottom=-5, top=-1)
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs

    assert res.vertical.values.size > 0
    assert np.all((res.vertical.values >= 1) & (res.vertical.values <= 5))


def test_geobbox_curvilinear_grid_vertical_partial(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.geobbox(top=-2.5)
    assert res.latitude.name == "latitude"
    assert res.longitude.name == "longitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"
    assert res.domain.crs == vt.domain.crs
    assert res.vertical.values.size > 0
    assert np.all(res.vertical.values >= 2.5)


def test_locations_curvilinear_grid_horizontal_1(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    lat, lon = -20, -115
    res = vt.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)

    assert res.vertical.type.name == "INDEPENDENT"
    assert res.vertical.shape == vt.vertical.shape
    assert np.allclose(res.vertical.values, vt.vertical.values)


def test_locations_curvilinear_grid_horizontal_2(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    lat, lon = [-20, -22], [-115, -120]
    res = vt.locations(latitude=lat, longitude=lon)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)

    assert res.vertical.type.name == "INDEPENDENT"
    assert res.vertical.shape == vt.vertical.shape
    assert np.allclose(res.vertical.values, vt.vertical.values)


def test_locations_curvilinear_grid_all_1(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    lat, lon, vert = -20, -115, -5
    res = vt.locations(latitude=lat, longitude=lon, vertical=vert)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)

    assert res.vertical.type.name == "DEPENDENT"
    assert len(res.vertical.dims) == 1
    assert res.vertical.dims[0].name == "points"
    assert np.allclose(
        -res.vertical.values, vert, atol=np.diff(vt.vertical.values).max() / 2
    )


def test_locations_curvilinear_grid_all_2(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    lat, lon, vert = [-20, -22], [-115, -120], [-5, -10]
    res = vt.locations(latitude=lat, longitude=lon, vertical=vert)

    assert res.latitude.name == "latitude"
    assert res.latitude.ncvar == "nav_lat"
    assert res.longitude.name == "longitude"
    assert res.longitude.ncvar == "nav_lon"
    assert res.vertical.name == "vertical"
    assert res.vertical.ncvar == "depthv"

    assert res.latitude.type.name == "DEPENDENT"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert np.allclose(res.latitude.values, lat, atol=0.1)

    assert res.longitude.type.name == "DEPENDENT"
    assert len(res.longitude.dims) == 1
    assert res.longitude.dims[0].name == "points"
    assert np.allclose(res.longitude.values, lon, atol=0.1)

    assert res.vertical.type.name == "DEPENDENT"
    assert len(res.vertical.dims) == 1
    assert res.vertical.dims[0].name == "points"
    assert np.allclose(
        -res.vertical.values, vert, atol=np.diff(vt.vertical.values).max() / 2
    )


def test_timecombo_single_hour(era5_netcdf):
    tp = Field.from_xarray(era5_netcdf, ncvar="tp")
    res = tp.sel(time={"year": 2020, "day": [1, 6, 10], "hour": 5})
    dset = res.to_xarray(True)
    assert np.all(
        (dset.time.dt.day == 1)
        | (dset.time.dt.day == 6)
        | (dset.time.dt.day == 10)
    )
    assert np.all(dset.time.dt.hour == 5)
    assert np.all(dset.time.dt.month == 6)
    assert np.all(dset.time.dt.year == 2020)


def test_timecombo_single_day(era5_netcdf):
    tp = Field.from_xarray(era5_netcdf, ncvar="tp")
    res = tp.sel(time={"year": 2020, "day": 10, "hour": [22, 5, 4]})
    dset = res.to_xarray(True)
    assert np.all(
        (dset.time.dt.hour == 4)
        | (dset.time.dt.hour == 5)
        | (dset.time.dt.hour == 22)
    )
    assert np.all(dset.time.dt.day == 10)
    assert np.all(dset.time.dt.month == 6)
    assert np.all(dset.time.dt.year == 2020)


# TODO: verify!
@pytest.mark.skip(
    "Should lat and lon depend on points if crs is RegularLatLon?"
)
def test_locations_regular_latlon_single_lat_multiple_lon(era5_netcdf):
    d2m = Field.from_xarray(era5_netcdf, ncvar="d2m")

    res = d2m.locations(latitude=[41, 41], longitude=[9, 12])
    assert res["latitude"].type is CoordinateType.INDEPENDENT
    assert res["longitude"].type is CoordinateType.INDEPENDENT
    assert np.all(res.latitude.values == 41)
    assert np.all((res.longitude.values == 9) | (res.longitude.values == 12))

    dset = res.to_xarray()
    assert np.all(dset.latitude == 41)
    assert np.all((dset.longitude == 9) | (dset.longitude == 12))
    assert dset["d2m"].attrs["units"] == "K"
    coords = dset["d2m"].attrs.get(
        "coordinates", dset["d2m"].encoding.get("coordinates")
    )
    assert "latitude" in coords


# TODO: verify!
@pytest.mark.skip(
    f"Lat depends on `points` but is single-element and should be SCALAR not"
    f" DEPENDENT"
)
def test_locations_regular_latlon_single_lat_single_lon(era5_netcdf):
    d2m = Field.from_xarray(era5_netcdf, ncvar="d2m")
    res = d2m.locations(latitude=41, longitude=9)
    assert np.all(res.latitude.values == 41)
    assert np.all(res.longitude.values == 9)
    assert res["latitude"].type is CoordinateType.SCALAR
    assert res["longitude"].type is CoordinateType.SCALAR


def test_locations_regular_latlon_single_point(era5_netcdf):
    d2m = Field.from_xarray(era5_netcdf, ncvar="d2m")

    res = d2m.locations(latitude=41, longitude=9)
    assert res.latitude.values.item() == 41
    assert res.longitude.values.item() == 9
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert res.latitude.type.name == "DEPENDENT"
    assert res.longitude.type.name == "DEPENDENT"
    assert res.domain.type.name == "POINTS"

    dset = res.to_xarray()
    assert dset.latitude.item() == 41
    assert dset.longitude.item() == 9
    assert dset["d2m"].attrs["units"] == "K"
    coords_str = dset["d2m"].attrs.get(
        "coordinates", dset["d2m"].encoding.get("coordinates")
    )
    assert set(coords_str.split(" ")) == {"latitude", "longitude"}
    assert "points" in res.to_xarray().dims
    assert "latitude" not in res.to_xarray().dims
    assert "longitude" not in res.to_xarray().dims


def test_locations_regular_latlon_multiple_lat_multiple_lon(era5_netcdf):
    d2m = Field.from_xarray(era5_netcdf, ncvar="d2m")

    res = d2m.locations(latitude=[41, 42], longitude=[9, 12])
    assert np.all((res.latitude.values == 41) | (res.latitude.values == 42))
    assert np.all((res.longitude.values == 9) | (res.longitude.values == 12))
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert len(res.latitude.dims) == 1
    assert res.latitude.dims[0].name == "points"
    assert res.latitude.type.name == "DEPENDENT"
    assert res.longitude.type.name == "DEPENDENT"
    assert res.domain.type.name == "POINTS"

    dset = res.to_xarray()
    assert np.all((dset.latitude == 41) | (dset.latitude == 42))
    assert np.all((dset.longitude == 9) | (dset.longitude == 12))
    assert dset["d2m"].attrs["units"] == "K"
    coords_str = dset["d2m"].attrs.get(
        "coordinates", dset["d2m"].encoding.get("coordinates")
    )
    assert set(coords_str.split(" ")) == {"latitude", "longitude"}
    assert "points" in res.to_xarray().dims
    assert "latitude" not in res.to_xarray().dims
    assert "longitude" not in res.to_xarray().dims


@pytest.mark.skip(
    "`as_cartopy_crs` is not implemented for NEMO CurvilinearGrid"
)
def test_locations_curvilinear_grid_multiple_lat_multiple_lon(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")

    res = vt.locations(latitude=[-20, -21], longitude=[-111, -114])

    dset = res.to_xarray(True)
    assert "points" in dset.dims
    assert dset.points.shape == (2,)
    assert np.all(
        (dset.nav_lat.values + 20 < 0.2) | (dset.nav_lat.values + 21 < 0.2)
    )
    assert np.all(
        (dset.nav_lon.values + 111 < 0.2) | (dset.nav_lon.values + 114 < 0.2)
    )
    assert dset.vt.attrs["units"] == "degree_C m/s"
    assert "coordinates" not in dset.vt.attrs

@pytest.mark.skip(
    "Skipping test"
)
def test_sel_fail_on_missing_x_y(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    with pytest.raises(KeyError, match=r"Axis of type*"):
        _ = vt.sel(depth=[1.2, 29], x=slice(60, 100), y=slice(130, 170))


def test_nemo_sel_vertical_fail_on_missing_value_if_method_undefined(
    nemo_ocean_16,
):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    with pytest.raises(KeyError):
        _ = vt.sel(depth=[1.2, 29])


def test_nemo_sel_vertical_with_std_name(nemo_ocean_16):
    nemo_ocean_16.depthv.attrs["standard_name"] = "vertical"
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    res = vt.sel(depth=[1.2, 29], method="nearest")
    assert len(res.vertical) == 2
    assert np.all(
        (res["vertical"].values == nemo_ocean_16.depthv.values[1])
        | (res["vertical"].values == nemo_ocean_16.depthv.values[-2])
    )


@pytest.mark.skipif(
    xr.__version__ > "2022.1.0",
    reason="Subsetting does not support empty indexers",
)
def test_nemo_sel_time_empty_result(nemo_ocean_16):
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    res = vt.sel(time={"hour": 13}, method="nearest")
    assert res["time"].shape == (0,)


def test_rotated_pole_sel_rlat_rlon_with_axis(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    res = wso.sel(y=slice(-1.7, -0.9), x=slice(4, 4.9))
    assert len(res[Axis("x")]) > 0
    assert len(res[Axis("y")]) > 0
    assert np.all(res[Axis("y")].values >= -1.7)
    assert np.all(res[Axis("y")].values <= -0.9)
    assert np.all(res[Axis("x")].values >= 4.0)
    assert np.all(res[Axis("x")].values <= 4.9)


def test_rotated_pole_sel_rlat_rlon_with_std_name(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    res = wso.sel(
        grid_latitude=slice(-1.7, -0.9), grid_longitude=slice(4, 4.9)
    )
    assert len(res[Axis("rlat")]) > 0
    assert len(res[Axis("rlon")]) > 0
    assert np.all(res[Axis("rlat")].values >= -1.7)
    assert np.all(res[Axis("rlat")].values <= -0.9)
    assert np.all(res[Axis("rlon")].values >= 4.0)
    assert np.all(res[Axis("rlon")].values <= 4.9)

    assert np.all(res[Axis("y")].values >= -1.7)
    assert np.all(res[Axis("y")].values <= -0.9)
    assert np.all(res[Axis("x")].values >= 4.0)
    assert np.all(res[Axis("x")].values <= 4.9)

@pytest.mark.skip(
    "Skipping test"
)
def test_rotated_pole_sel_lat_with_std_name_fails(era5_rotated_netcdf):
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    with pytest.raises(KeyError):
        _ = wso.sel(latitude=slice(39, 41))


def test_rotated_pole_sel_time_with_diff_ncvar(era5_rotated_netcdf):
    era5_rotated_netcdf = era5_rotated_netcdf.rename({"time": "tm"})
    wso = Field.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    res = wso.sel(time=slice("2007-05-02T00:00:00", "2007-05-02T11:00:00"))
    assert len(res.time) > 0
    assert np.all(
        res[Axis("time")].values <= np.datetime64("2007-05-02T11:00:00")
    )
    assert np.all(
        res[Axis("time")].values >= np.datetime64("2007-05-02T00:00:00")
    )
    assert np.all(res.time.values <= np.datetime64("2007-05-02T11:00:00"))
    assert np.all(res.time.values >= np.datetime64("2007-05-02T00:00:00"))
    assert np.all(res["time"].values <= np.datetime64("2007-05-02T11:00:00"))
    assert np.all(res["time"].values >= np.datetime64("2007-05-02T00:00:00"))

    dset = res.to_xarray(encoding=True)
    assert "time" not in dset.coords
    assert "tm" in dset.coords

@pytest.mark.skip(
    "Invalidate as in the current version, bounds aren't computed correctly"
)
def test_nemo_sel_proper_ncvar_name_in_res(nemo_ocean_16):
    nemo_ocean_16["vt"].attrs["standard_name"] = "vt_std_name"
    vt = Field.from_xarray(nemo_ocean_16, ncvar="vt")
    assert vt.name == "vt_std_name"
    assert vt.ncvar == "vt"
    res = vt.sel(depth=1.2, method="nearest")
    assert res.name == "vt_std_name"
    assert res.ncvar == "vt"

    dset = res.to_xarray(encoding=True)
    assert "vt" in dset.data_vars
    assert "vt_std_name" not in dset.data_vars
    assert "nav_lat" in dset.coords
    assert "latitude" not in dset.coords
    assert "bounds_lat" in dset.coords
    assert "bounds_lon" in dset.coords

    dset = res.to_xarray(encoding=False)
    assert "vt" not in dset.data_vars
    assert "vt_std_name" in dset.data_vars
    assert "nav_lat" not in dset.coords
    assert "latitude" in dset.coords
    assert "bounds_lat" in dset.coords
    assert "bounds_lon" in dset.coords


def test_field_create_with_dict_coords():
    dims = ("time", Axis("latitude"), AxisType.LONGITUDE)
    coords = {
        "time": pd.date_range("06-06-2019", "19-12-2019", periods=50),
        AxisType.LATITUDE: np.linspace(15, 100, 40),
        AxisType.LONGITUDE: np.linspace(5, 10, 30),
    }
    f = Field(
        name="ww",
        data=np.random.random((50, 40, 30)),
        dims=dims,
        coords=coords,
        units="m",
        encoding={"name": "w_ncvar"},
    )
    assert f.name == "ww"
    assert f.ncvar == "w_ncvar"
    assert f.dim_names == ("time", "latitude", "longitude")
    assert f.domain.crs == RegularLatLon()
    assert np.all(
        f[Axis("time")].values
        == np.array(pd.date_range("06-06-2019", "19-12-2019", periods=50))
    )
    assert np.all(
        f[AxisType.TIME].values
        == np.array(pd.date_range("06-06-2019", "19-12-2019", periods=50))
    )
    assert np.all(f[AxisType.LATITUDE].values == np.linspace(15, 100, 40))
    assert np.all(f[Axis("lat")].values == np.linspace(15, 100, 40))
    assert np.all(f[Axis("lon")].values == np.linspace(5, 10, 30))
    assert np.all(f[Axis("longitude")].values == np.linspace(5, 10, 30))
    assert np.all(f[AxisType.LONGITUDE].values == np.linspace(5, 10, 30))
    assert f.units._unit == cf.Unit("m")


def test_var_name_when_field_from_field_id_is_missing(era5_rotated_netcdf):
    wso = Field.from_xarray(
        era5_rotated_netcdf,
        ncvar="W_SO",
        id_pattern="{standard_name}_{not_existing_fied}",
    )
    assert wso.name == "W_SO"
    assert wso.latitude.name == "lat"
    assert wso.longitude.name == "lon"
    assert wso.time.name == "time"

    wso = Field.from_xarray(
        era5_rotated_netcdf, ncvar="W_SO", id_pattern="{standard_name}"
    )
    assert wso.name == "lwe_thickness_of_moisture_content_of_soil_layer"
    assert wso.latitude.name == "latitude"
    assert wso.longitude.name == "longitude"
    assert wso.time.name == "time"

@pytest.mark.skip(
    "Invalidate as in the current version, bounds aren't computed correctly"
)
def test_to_xarray_time_with_bounds(era5_rotated_netcdf, nemo_ocean_16):
    field = Field.from_xarray(era5_rotated_netcdf, "W_SO")
    da = field.to_xarray(encoding=False)
    assert "bounds" in da["time"].encoding
    assert da["time"].encoding["bounds"] == "time_bnds"
    assert "time_bnds" in da.coords

    da = field.to_xarray(encoding=True)
    assert "bounds" in da["time"].encoding
    assert da["time"].encoding["bounds"] == "time_bnds"
    assert "time_bnds" in da.coords

    field = Field.from_xarray(nemo_ocean_16, "vt")
    da = field.to_xarray(encoding=False)
    assert "bounds" in da["time"].encoding
    assert da["time"].encoding["bounds"] == "time_counter_bounds"
    assert "time_counter_bounds" in da.coords

    da = field.to_xarray(encoding=True)
    assert "bounds" in da["time_counter"].encoding
    assert da["time_counter"].encoding["bounds"] == "time_counter_bounds"
    assert "time_counter_bounds" in da.coords

@pytest.mark.skip(
    "Invalidate as in the current version, bounds aren't computed correctly"
)
def test_to_xarray_time_with_bounds_mapping(era5_rotated_netcdf):
    field = Field.from_xarray(
        era5_rotated_netcdf,
        "W_SO",
        mapping={"time_bnds": {"name": "time_bounds_name"}},
    )
    da = field.to_xarray(encoding=False)
    assert "bounds" in da["time"].encoding
    assert da["time"].encoding["bounds"] == "time_bounds_name"
    assert "time_bounds_name" in da.coords

@pytest.mark.skip(
    "Invalidate as in the current version, bounds aren't computed correctly"
)
def test_to_xarray_time_with_bounds_nemo_with_mapping(nemo_ocean_16):
    field = Field.from_xarray(
        nemo_ocean_16,
        "vt",
        mapping={"time_counter_bounds": {"name": "time_bnds"}},
    )
    da = field.to_xarray(encoding=False)
    assert "bounds" in da["time"].encoding
    assert da["time"].encoding["bounds"] == "time_bnds"
    assert "time_bnds" in da.coords

    da = field.to_xarray(encoding=True)
    assert "bounds" in da["time_counter"].encoding
    assert da["time_counter"].encoding["bounds"] == "time_bnds"
    assert "time_bnds" in da.coords


def test_geobbox_regular_latlon_negative_convention(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    res = tp.geobbox(north=10, south=-10, west=25, east=25)
    assert np.all(res.longitude <= 25)
    assert np.all(res.longitude >= 25)
    assert np.all(res.latitude <= 10)
    assert np.all(res.latitude >= -10)


def test_sel_longitude_with_leftnone_slice(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_longitude = tp.sel(longitude=slice(None, 20)).longitude.values
    assert np.all(selected_longitude <= 20)
    assert np.all(selected_longitude >= np.min(tp.longitude.values))


def test_sel_longitude_with_rightnone_slice(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_longitude = tp.sel(longitude=slice(45, None)).longitude.values
    assert np.all(selected_longitude <= np.max(tp.longitude.values))
    assert np.all(selected_longitude >= 45)


def test_sel_longitude_with_leftnone_slice_change_convention(
    era5_globe_netcdf,
):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_longitude = tp.sel(longitude=slice(None, -10)).longitude.values
    assert np.all(selected_longitude <= -10)
    assert np.all(selected_longitude >= -180)


def test_sel_longitude_with_rightnone_slice_change_convention(
    era5_globe_netcdf,
):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_longitude = tp.sel(longitude=slice(-20, None)).longitude.values
    assert np.all(selected_longitude <= 180)
    assert np.all(selected_longitude >= -20)


def test_sel_latitude_with_leftnone_slice(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_latitude = tp.sel(latitude=slice(None, 20)).latitude.values
    assert np.all(selected_latitude >= 20)
    assert np.all(selected_latitude <= np.max(tp.latitude.values))


def test_sel_latitude_with_rightnone_slice(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    selected_latitude = tp.sel(latitude=slice(45, None)).latitude.values
    assert np.all(selected_latitude >= np.min(tp.latitude.values))
    assert np.all(selected_latitude <= 45)


def test_interpolate_regular_to_regular_gridded(era5_netcdf):
    field = Field.from_xarray(era5_netcdf["d2m"].to_dataset(), ncvar="d2m")
    loc = {
        "latitude": np.linspace(35, 48, num=20),
        "longitude": np.linspace(2, 20, num=20),
    }
    domain = GeodeticGrid(**loc)
    domain.type = DomainType.GRIDDED
    res = field.interpolate(domain=domain, method="linear")
    assert res.domain.crs == domain.crs
    assert res.domain.type == domain.type
    dims = {dim.type for dim in res.dims.flat}
    assert AxisType.LATITUDE in dims
    assert AxisType.LONGITUDE in dims
    assert res.latitude.size == loc["latitude"].size
    assert res.longitude.size == loc["longitude"].size


def test_interpolate_regular_to_regular_point(era5_netcdf):
    field = Field.from_xarray(era5_netcdf["d2m"].to_dataset(), ncvar="d2m")
    loc = {
        "latitude": np.linspace(35, 48, num=20),
        "longitude": np.linspace(2, 20, num=20),
    }
    domain = GeodeticPoints(**loc)
    res = field.interpolate(domain=domain, method="linear")
    assert res.domain.crs == domain.crs
    assert res.domain.type == domain.type
    assert res.latitude.dims.size == 1
    assert res.latitude.dims[0].name == "points"
    assert res.latitude.size == loc["latitude"].size
    assert res.longitude.dims.size == 1
    assert res.longitude.dims[0].name == "points"
    assert res.longitude.size == loc["longitude"].size


def test_interpolate_rotated_pole_to_regular_gridded(era5_rotated_netcdf_wso):
    wso = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    loc = {
        "latitude": np.linspace(35, 48, num=20),
        "longitude": np.linspace(2, 20, num=20),
    }
    domain = GeodeticGrid(**loc)
    domain.type = DomainType.GRIDDED
    res = wso.interpolate(domain=domain, method="linear")
    assert res.domain.crs == domain.crs
    assert res.domain.type == domain.type
    dims = {dim.type for dim in res.dims.flat}
    assert AxisType.LATITUDE in dims
    assert AxisType.LONGITUDE in dims
    assert res.latitude.size == loc["latitude"].size
    assert res.longitude.size == loc["longitude"].size


def test_interpolate_rotated_pole_to_regular_point(era5_rotated_netcdf_wso):
    wso = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    loc = {
        "latitude": np.linspace(35, 48, num=20),
        "longitude": np.linspace(2, 20, num=20),
    }
    domain = GeodeticPoints(**loc)
    res = wso.interpolate(domain=domain, method="linear")
    assert res.domain.crs == domain.crs
    assert res.domain.type == domain.type
    assert res.latitude.dims.size == 1
    assert res.latitude.dims[0].name == "points"
    assert res.latitude.size == loc["latitude"].size
    assert res.longitude.dims.size == 1
    assert res.longitude.dims[0].name == "points"
    assert res.longitude.size == loc["longitude"].size


def test_resample_without_original_bounds(era5_globe_netcdf):
    tp = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    tp_r = tp.resample(operator="maximum", frequency="W")
    assert tp.values.max() == tp_r.values.max()
    diff = 7 * 24 * 60 * 60 * 1_000_000_000
    diff_ = np.diff(tp_r.time.values)
    assert np.all(diff_ == np.full_like(diff_, fill_value=diff))
    assert tp_r.time.bounds is not None
    assert tp_r.time.bounds["time_bounds"].shape[0] == tp_r.time.shape[0]


def test_resample_with_original_bounds(era5_rotated_netcdf_wso):
    wso = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    wso_r = wso.resample(operator="sum", frequency="12H")
    diff = 12 * 60 * 60 * 1_000_000_000
    diff_ = np.diff(wso_r.time.values)
    assert np.all(diff_ == np.full_like(diff_, fill_value=diff))
    assert wso_r.time.bounds is not None
    assert wso_r.time.bounds["time_bounds"].shape[0] == wso_r.time.shape[0]


def test_adding_time_bounds(era5_netcdf):
    field = Field.from_xarray(era5_netcdf["d2m"].to_dataset(), ncvar="d2m")
    assert field.time.bounds is None
    time_bounds = np.empty(shape=(field.time.size, 2), dtype=field.time.dtype)
    time_resolution = field.time.values[1] - field.time.values[0]
    time_bounds[0, 0] = field.time.values[0] - time_resolution
    time_bounds[1:, 0] = field.time.values[:-1]
    time_bounds[:, 1] = field.time.values
    field.time.bounds = time_bounds
    assert isinstance(field.time.bounds, dict)
    assert "time_bounds" in field.time.bounds
    assert field.time.bounds["time_bounds"].shape == field.time.shape + (2,)
    assert "bounds" in field.time.encoding
    assert field.time.encoding["bounds"] == "time_bounds"

@pytest.mark.skip(
    "Invalidated test as in the current version bound are not computed correctly"
)
def test_regridding_regular_to_regular_conservative(era5_netcdf):
    field_in = Field.from_xarray(era5_netcdf, ncvar="d2m")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="conservative",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    assert field_out.coords.keys() == {"time", "latitude", "longitude"}
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_regular_to_regular_bilinear(era5_netcdf):
    field_in = Field.from_xarray(era5_netcdf, ncvar="d2m")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="bilinear",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    assert field_out.coords.keys() == {"time", "latitude", "longitude"}
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_regular_to_regular_nearest_s2d(era5_netcdf):
    field_in = Field.from_xarray(era5_netcdf, ncvar="d2m")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="nearest_s2d",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    assert field_out.coords.keys() == {"time", "latitude", "longitude"}
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_regular_to_regular_nearest_d2s(era5_netcdf):
    field_in = Field.from_xarray(era5_netcdf, ncvar="d2m")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="nearest_d2s",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    assert field_out.coords.keys() == {"time", "latitude", "longitude"}
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_regular_to_regular_patch(era5_netcdf):
    field_in = Field.from_xarray(era5_netcdf, ncvar="d2m")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="patch",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    assert field_out.coords.keys() == {"time", "latitude", "longitude"}
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_rotated_pole_to_regular_bilinear(era5_rotated_netcdf_wso):
    field_in = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="bilinear",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    dims = field_in.coords.keys() - {"grid_latitude", "grid_longitude"}
    assert field_out.coords.keys() == dims
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_rotated_pole_to_regular_nearest_s2d(
    era5_rotated_netcdf_wso,
):
    field_in = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="nearest_s2d",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    dims = field_in.coords.keys() - {"grid_latitude", "grid_longitude"}
    assert field_out.coords.keys() == dims
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_rotated_pole_to_regular_nearest_d2s(
    era5_rotated_netcdf_wso,
):
    field_in = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="nearest_d2s",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    dims = field_in.coords.keys() - {"grid_latitude", "grid_longitude"}
    assert field_out.coords.keys() == dims
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )

def test_regridding_rotated_pole_to_regular_patch(era5_rotated_netcdf_wso):
    field_in = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    lat_in, lon_in = field_in.latitude, field_in.longitude

    lat_step = lon_step = 0.1
    lat = np.arange(lat_in.min(), lat_in.max(), lat_step)
    lon = np.arange(lon_in.min(), lon_in.max(), lon_step)
    target = Domain(
        coords=[
            Coordinate(
                data=lat,
                axis=Axis(name="latitude", is_dim=True),
                dims=("latitude",),
            ),
            Coordinate(
                data=lon,
                axis=Axis(name="longitude", is_dim=True),
                dims=("longitude",),
            ),
        ],
        crs=GeogCS(6371229),
        domaintype=DomainType.GRIDDED,
    )
    field_out = field_in.regrid(
        target=target,
        method="patch",
        weights_path=None,
        reuse_weights=True,
    )

    assert field_out.domain.crs == target.crs
    dims = field_in.coords.keys() - {"grid_latitude", "grid_longitude"}
    assert field_out.coords.keys() == dims
    assert field_in.time.shape == field_out.time.shape
    assert np.all(field_in.time.values == field_out.time.values)
    assert target.latitude.shape == field_out.latitude.shape
    assert np.allclose(target.latitude.values, field_out.latitude.values)
    assert target.longitude.shape == field_out.longitude.shape
    assert np.allclose(target.longitude.values, field_out.longitude.values)
    assert np.allclose(
        np.nanquantile(field_in.values, [0, 0.25, 0.5, 0.75, 1]),
        np.nanquantile(field_out.values, [0, 0.25, 0.5, 0.75, 1]),
        atol=1,
    )


def test_sel_by_time_combo_only_day(era5_netcdf):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    field = field.sel(time={"day": 10})
    assert len(field.time) == 24


def test_resample_if_gap_in_time_axis(era5_netcdf):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    field = field.sel(time={"day": [10, 15], "hour": [10, 16]})
    val1 = (
        field.to_xarray(False)
        .tp.isel(time=field.time.to_xarray(False).time.dt.day.values == 10)
        .sum(dim=field.time.name)
    )
    val2 = (
        field.to_xarray(False)
        .tp.isel(time=field.time.to_xarray(False).time.dt.day.values == 15)
        .sum(dim=field.time.name)
    )
    result = field.resample(frequency="1D", operator="sum")
    assert len(result.time) == 2
    assert np.allclose(result.to_xarray(False).tp.values[0, ...], val1)
    assert np.allclose(result.to_xarray(False).tp.values[1, ...], val2)


def test_auxiliary_coords_after_resampling(era5_rotated_netcdf_tmin2m):
    field = Field.from_xarray(era5_rotated_netcdf_tmin2m, ncvar="TMIN_2M")
    assert "latitude" in field.domain.coords
    assert field.domain.coords["latitude"].type is CoordinateType.DEPENDENT
    assert "longitude" in field.domain.coords
    assert field.domain.coords["longitude"].type is CoordinateType.DEPENDENT
    field = field.resample(frequency="1D", operator="min")
    assert "latitude" in field.domain.coords
    assert field.domain.coords["latitude"].type is CoordinateType.DEPENDENT
    assert "longitude" in field.domain.coords
    assert field.domain.coords["longitude"].type is CoordinateType.DEPENDENT


@pytest.mark.skip("Currently domaintype is set only in `locations` method")
def test_field_domain_type_regular_lat_lon(
    era5_netcdf,
):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    assert field.domain.type is DomainType.GRIDDED


def test_keeping_field_domain_type_in_to_xarray_and_from_xarray(era5_netcdf):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    field = field.locations(latitude=10, longitude=20)
    assert field.domain.type is DomainType.POINTS
    field = field.sel(time={"day": 10})
    assert field.domain.type is DomainType.POINTS


def test_field_serialization_success_with_domtype_attr(era5_netcdf):
    field = Field.from_xarray(era5_netcdf, ncvar="tp")
    field = field.locations(latitude=10, longitude=20)
    assert field.domain.type is DomainType.POINTS
    field.to_netcdf(RES_PATH)
    clear_test_res()


def test_extract_polygons_regular(era5_globe_netcdf, it_shape_100_km):
    tp = era5_globe_netcdf["tp"]
    field = Field.from_xarray(era5_globe_netcdf, ncvar="tp")
    coord_names = field.coords.keys()
    field, mask = field.extract_polygons(
        geometry=it_shape_100_km, crop=True, return_mask=True
    )
    lon_lb = it_shape_100_km.bounds["minx"].min()
    lon_ub = it_shape_100_km.bounds["maxx"].max()
    lat_lb = it_shape_100_km.bounds["miny"].min()
    lat_ub = it_shape_100_km.bounds["maxy"].max()
    assert mask.values.dtype == np.bool_
    assert mask.coords.keys() == {"latitude", "longitude"}
    assert np.allclose(mask.coords["latitude"], tp.coords["latitude"])
    assert np.allclose(mask.coords["longitude"], tp.coords["longitude"])
    assert field.coords.keys() == coord_names
    assert field.latitude.values.min() >= lat_lb
    assert field.latitude.values.max() <= lat_ub
    assert field.longitude.values.min() >= lon_lb
    assert field.longitude.values.max() <= lon_ub
    assert np.nanmin(field.values) >= np.nanmin(tp)
    assert np.nanmin(field.values) <= np.nanmax(tp)


def test_extract_polygons_rotated_pole(
    era5_rotated_netcdf_wso, it_shape_100_km
):
    wso = era5_rotated_netcdf_wso["W_SO"]
    field = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    field, mask = field.extract_polygons(
        geometry=it_shape_100_km, crop=True, return_mask=True
    )
    lon_lb = it_shape_100_km.bounds["minx"].min()
    lon_ub = it_shape_100_km.bounds["maxx"].max()
    lat_lb = it_shape_100_km.bounds["miny"].min()
    lat_ub = it_shape_100_km.bounds["maxy"].max()
    assert mask.values.dtype == np.bool_
    assert mask.coords.keys() == {"latitude", "longitude"}
    assert field.latitude.values.min() >= lat_lb
    assert field.latitude.values.max() <= lat_ub
    assert field.longitude.values.min() >= lon_lb
    assert field.longitude.values.max() <= lon_ub
    assert np.nanmin(field.values) >= np.nanmin(wso)
    assert np.nanmin(field.values) <= np.nanmax(wso)


def test_using_geo_domtype_attribute_in_serialization_field(
    era5_rotated_netcdf_wso,
):
    clear_test_res()
    field = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    field.domain.type = DomainType("points")
    field_dset = field.to_xarray(False)[
        "lwe_thickness_of_moisture_content_of_soil_layer"
    ]
    assert "__geo_domtype" in field_dset.attrs
    assert isinstance(field_dset.attrs["__geo_domtype"], DomainType)
    with pytest.raises(TypeError, match=r"Invalid value for attr*"):
        field_dset.to_netcdf(RES_PATH)

    field_dset = field.to_xarray(True)["W_SO"]
    assert "__geo_domtype" not in field_dset.attrs
    field_dset.to_netcdf(RES_PATH)
    clear_test_res()


def test_keep_encoding_after_to_regular(era5_rotated_netcdf_wso):
    field = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    reg_field = field.to_regular()
    compare_dicts(
        field.encoding,
        reg_field.encoding,
        exclude_d1=["grid_mapping", "coordinates", "_FillValue"],
        exclude_d2=["grid_mapping", "coordinates", "_FillValue"],
    )


def test_keep_coordinate_encoding_with_scalars_after_to_regular(
    era5_rotated_netcdf_wso,
):
    era5_rotated_netcdf_wso = era5_rotated_netcdf_wso.assign_coords({"sc": 10})
    era5_rotated_netcdf_wso["sc"].attrs["standard_name"] = "scalar_coord"
    era5_rotated_netcdf_wso["W_SO"].encoding["coordinates"] = "lat lon sc"
    field = Field.from_xarray(era5_rotated_netcdf_wso, ncvar="W_SO")
    reg_field = field.to_regular()
    xr_ds = reg_field.to_xarray()
    assert xr_ds["W_SO"].encoding["coordinates"] == "sc"



import os
import cartopy.crs as ccrs
import numpy as np
import pytest
import xarray as xr

import geokube.core.coord_system as crs
from geokube.backend import open_datacube
from geokube.core.axis import Axis, AxisType
from geokube.core.coord_system import RegularLatLon
from geokube.core.coordinate import Coordinate
from geokube.core.datacube import DataCube
from geokube.core.field import Field
from geokube.core.domain import Domain
from geokube.core.unit import Unit
from geokube.core.variable import Variable

from tests import RES_PATH, RES_DIR, clear_test_res
from tests.fixtures import *


def test_from_xarray_regular_lat_lon(era5_netcdf):
    dc = DataCube.from_xarray(era5_netcdf)
    assert "tp" in dc
    assert "d2m" in dc
    assert dc.properties == era5_netcdf.attrs
    assert dc.encoding == era5_netcdf.encoding


def test_from_xarray_regular_lat_lon_with_id_pattern(era5_netcdf):
    dc = DataCube.from_xarray(era5_netcdf, id_pattern="{__ddsapi_name}")
    assert "total_precipitation" in dc
    assert "2_metre_dewpoint_temperature" in dc
    assert dc["total_precipitation"].domain.crs == crs.RegularLatLon()
    assert dc["2_metre_dewpoint_temperature"].domain.crs == crs.RegularLatLon()

    assert dc["2_metre_dewpoint_temperature"].units == Unit("K")
    assert dc["total_precipitation"].units == Unit("m")


def test_to_xarray_regular_lat_lon_with_id_pattern_without_encoding(
    era5_netcdf,
):
    dc = DataCube.from_xarray(era5_netcdf, id_pattern="{__ddsapi_name}")
    xr_res = dc.to_xarray(encoding=False)
    assert "2_metre_dewpoint_temperature" in xr_res.data_vars
    assert "total_precipitation" in xr_res.data_vars
    assert "crs_latitude_longitude" in xr_res.coords


def test_to_xarray_regular_lat_lon_with_id_pattern_with_encoding(era5_netcdf):
    dc = DataCube.from_xarray(era5_netcdf, id_pattern="{__ddsapi_name}")
    xr_res = dc.to_xarray(encoding=True)
    assert "d2m" in xr_res.data_vars
    assert "tp" in xr_res.data_vars
    assert "crs_latitude_longitude" in xr_res.coords


def test_geobbox_regular_latlon(era5_globe_netcdf):
    dc = DataCube.from_xarray(era5_globe_netcdf)
    res = dc.geobbox(north=10, south=-10, west=-20, east=20)
    assert np.all(res["tp"]["latitude"].values <= 10)
    assert np.all(res["tp"]["latitude"].values >= -10)
    assert np.all(res["tp"].latitude.values <= 10)
    assert np.all(res["tp"].latitude.values >= -10)

    assert np.all(res["tp"]["longitude"].values <= 20)
    assert np.all(res["tp"]["longitude"].values >= -20)
    assert np.all(res["tp"].longitude.values <= 20)
    assert np.all(res["tp"].longitude.values >= -20)

    dset = res.to_xarray(True)
    assert "tp" in dset
    assert np.all(dset.latitude <= 10)
    assert np.all(dset.latitude >= -10)
    assert dset.latitude.attrs["units"] == "degrees_north"

    assert np.all(dset.longitude <= 20)
    assert np.all(dset.longitude >= -20)
    assert dset.longitude.attrs["units"] == "degrees_east"


def test_geobbox_rotated_pole(era5_rotated_netcdf):
    wso = DataCube.from_xarray(era5_rotated_netcdf)

    res = wso.geobbox(north=40, south=38, west=16, east=19)
    assert res[
        "lwe_thickness_of_moisture_content_of_soil_layer"
    ].domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    assert res["air_temperature"].domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    W = np.prod(
        res["lwe_thickness_of_moisture_content_of_soil_layer"].latitude.shape
    )
    assert (
        np.sum(
            res[
                "lwe_thickness_of_moisture_content_of_soil_layer"
            ].latitude.values
            >= 38
        )
        / W
        > 0.95
    )
    assert (
        np.sum(
            res[
                "lwe_thickness_of_moisture_content_of_soil_layer"
            ].latitude.values
            <= 40
        )
        / W
        > 0.95
    )
    assert (
        np.sum(
            res[
                "lwe_thickness_of_moisture_content_of_soil_layer"
            ].longitude.values
            >= 16
        )
        / W
        > 0.95
    )
    assert (
        np.sum(
            res[
                "lwe_thickness_of_moisture_content_of_soil_layer"
            ].longitude.values
            <= 19
        )
        / W
        > 0.95
    )

    dset = res.to_xarray(encoding=True)
    assert "TMIN_2M" in dset.data_vars
    assert "W_SO" in dset.data_vars
    assert "rlat" in dset.coords
    assert "rlon" in dset.coords
    assert "lat" in dset
    assert dset.lat.attrs["units"] == "degrees_north"
    assert "lon" in dset
    assert dset.lon.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords

    dset = res.to_xarray(False)
    assert "air_temperature" in dset.data_vars
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in dset.data_vars
    assert "latitude" in dset
    assert dset.latitude.attrs["units"] == "degrees_north"
    assert "longitude" in dset
    assert dset.longitude.attrs["units"] == "degrees_east"
    assert "crs_rotated_latitude_longitude" in dset.coords


def test_getitem_with_coords_and_fiels(era5_rotated_netcdf):
    dc = DataCube.from_xarray(era5_rotated_netcdf)
    assert "W_SO" in dc
    assert "TMIN_2M" in dc
    assert Axis("latitude") in dc
    assert "lat" in dc
    assert AxisType.TIME in dc
    assert "air_temperature" in dc
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in dc
    wso = dc["W_SO"]
    assert wso.name == "lwe_thickness_of_moisture_content_of_soil_layer"
    assert wso.ncvar == "W_SO"
    tmin = dc["TMIN_2M"]
    assert tmin.name == "air_temperature"
    assert tmin.ncvar == "TMIN_2M"
    lat = dc.latitude
    assert lat.name == "latitude"
    assert lat.ncvar == "lat"
    assert lat.axis_type is AxisType.LATITUDE
    time = dc.time
    assert time.ncvar == "time"
    assert time.axis_type is AxisType.TIME
    assert dc["latitude"] is not None
    assert dc.latitude is not None
    assert dc[AxisType.LATITUDE] is not None
    assert dc[Axis("lat")] is not None


def test_locations_rotated_pole(era5_rotated_netcdf):
    dc = DataCube.from_xarray(era5_rotated_netcdf)
    res = dc.locations(latitude=[39, 41], longitude=[17, 17])
    assert "points" in res.to_xarray().dims
    assert len(res.domain[AxisType.LATITUDE]) == 2
    assert len(res.domain[AxisType.LONGITUDE]) == 2
    assert res.domain.crs == RegularLatLon()


def test_get_multiple_fields_by_standard_name_list(rotated_pole_datacube):
    res = rotated_pole_datacube[
        ["lwe_thickness_of_moisture_content_of_soil_layer", "air_temperature"]
    ]
    assert len(res) == 2
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in res.fields
    assert "air_temperature" in res.fields


def test_get_single_field_by_standard_name(rotated_pole_datacube):
    res = rotated_pole_datacube["W_SO"]
    assert isinstance(res, Field)
    assert res.name == "lwe_thickness_of_moisture_content_of_soil_layer"
    assert res.ncvar == "W_SO"

    res = rotated_pole_datacube["TMIN_2M"]
    assert isinstance(res, Field)
    assert res.name == "air_temperature"
    assert res.ncvar == "TMIN_2M"


def test_persist_with_no_extension(rotated_pole_datacube):
    clear_test_res()
    with pytest.warns(UserWarning, match=r"Provided persistance path*"):
        path = rotated_pole_datacube.persist(RES_DIR)
        assert path == RES_DIR + ".nc"
    clear_test_res()


def test_persist_no_path_passed(rotated_pole_datacube):
    path = rotated_pole_datacube.persist()
    assert os.path.exists(path)


def test_to_dict_store_propen_keys(era5_netcdf):
    details = DataCube.from_xarray(era5_netcdf).to_dict()
    assert "domain" in details
    assert isinstance(details["domain"], dict)
    assert "fields" in details
    assert isinstance(details["fields"], dict)



import pytest

from geokube.core.axis import Axis, AxisType
from geokube.core.unit import Unit
from tests.fixtures import *


def test_init_proper_attributes_set():
    assert AxisType.LATITUDE.default_unit == Unit("degrees_north")
    assert AxisType.TIME.axis_type_name == "time"
    assert AxisType("aaa") is AxisType.GENERIC
    assert AxisType.GENERIC.default_unit == Unit("unknown")

    a1 = Axis(name="LAT", axistype="latitude")

    assert a1.name == "LAT"
    assert a1.type is AxisType.LATITUDE
    assert a1.encoding is None


def test_parsing():
    assert AxisType.parse("latitude") is AxisType.LATITUDE
    assert AxisType.parse("lat") is AxisType.LATITUDE
    assert AxisType.parse("rlat") is AxisType.Y
    assert AxisType.parse("x").default_unit == Unit("m")
    assert AxisType.parse("depth") is AxisType.VERTICAL
    assert AxisType.parse("time").default_unit == Unit(
        "hours since 1970-01-01", calendar="gregorian"
    )


def test_init_fails():
    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `str`,"
            r" `geokube.AxisType`, but provided *"
        ),
    ):
        _ = Axis("lon", axistype=10)

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `str`,"
            r" `geokube.AxisType`, but provided *"
        ),
    ):
        _ = Axis("lon", axistype={"lat"})

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `str`,"
            r" `geokube.AxisType`, but provided *"
        ),
    ):
        _ = Axis("lon", axistype=["lon"])


def test_init_from_other_axis():
    a1 = Axis(name="LAT", axistype="latitude")
    a3 = Axis(a1)
    assert id(a3) != id(a1)
    assert a3.name == a1.name
    assert a3.type is a1.type
    assert a3.encoding == a1.encoding
    assert a3.default_unit == Unit("degrees_north")

    a4 = Axis("lon", encoding={"name": "ncvar_my_name"})
    assert a4.name == "lon"
    assert a4.type is AxisType.LONGITUDE
    assert a4.encoding == {"name": "ncvar_my_name"}
    assert a4.default_unit == Unit("degrees_east")
    assert a4.ncvar == "ncvar_my_name"
    assert not a4.is_dim

    a5 = Axis(a4)
    assert a5.name == a4.name
    assert a5.type is a4.type
    assert a5.encoding == a4.encoding
    assert a5.default_unit == Unit("degrees_east")
    assert a5.ncvar == "ncvar_my_name"
    assert not a5.is_dim

    a6 = Axis("time", is_dim=True)
    assert a6.name == "time"
    assert a6.type is AxisType.TIME
    assert a6.is_dim


def test_axis_hash():
    a1 = Axis(AxisType.LONGITUDE)
    a2 = Axis(AxisType.LONGITUDE)
    assert a1 == a2
    assert hash(a1) == hash(a2)

    a1 = Axis("longitude")
    a2 = Axis(name="longitude", axistype=AxisType.LONGITUDE)
    assert a1 == a2
    assert hash(a1) == hash(a2)

    a1 = Axis("longitude", is_dim=True)
    a2 = Axis(AxisType.LONGITUDE, is_dim=False)
    assert a1 != a2
    assert hash(a1) != hash(a2)

    a1 = Axis("lat", AxisType.LONGITUDE)
    a2 = Axis(AxisType.LONGITUDE)
    assert a1 != a2
    assert hash(a1) != hash(a2)

    d = {Axis("lat"): [1, 2, 3], Axis("depth"): [-1, -2, -3]}
    assert Axis("lat") in d
    assert Axis("depth") in d
    assert d[Axis("lat")] == [1, 2, 3]
    assert d[Axis("depth")] == [-1, -2, -3]


def test_vertical_axis_pattern():
    assert Axis("dept").type is AxisType.VERTICAL
    assert Axis("depth").type is AxisType.VERTICAL



import warnings
from functools import wraps


def nonclose_assert(f):
    @wraps(f)
    def decorated_function(self, *args, **kwargs):
        if hasattr(self, "_is_closed") and self._is_closed:
            raise RuntimeError(
                "This object is already closed. Reinitialize it to do "
                "computations!"
            )
        return f(self, *args, **kwargs)

    return decorated_function


def geokube_logging(f):
    @wraps(f)
    def decorated_function(self, *args, **kwargs):
        if not hasattr(self, "_LOG"):
            warnings.warn(
                f"The class of the decorated `{str(f)}` does not contain"
                " `_LOG` object."
            )
            return f(self, *args, **kwargs)
        self._LOG.debug(f"Entering `{str(f)}`")
        try:
            return f(self, *args, **kwargs)
        except Exception as e:
            self._LOG.error(
                f"{type(e).__name__} with message `{e}` raised while executing"
                f" `{str(f)}`"
            )
            raise e
        finally:
            self._LOG.debug(f"Exiting `{str(f)}`")

    return decorated_function



import logging


class HCubeLogger:
    __slots__ = ("_logger",)

    def __init__(self, name):
        self._logger = logging.getLogger(name)

    def info(self, msg):
        self._logger.info(msg)

    def debug(self, msg):
        self._logger.debug(msg)

    def warn(self, msg):
        self._logger.warning(msg)

    def error(self, msg):
        self._logger.error(msg)



from __future__ import annotations

from typing import Iterable, Union

import numpy as np

from .enums import MethodType


class CellMethod:
    # Based on  https://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/build/ch07s03.html
    # Currently supports non-combined cell methods
    # TODO: extend for combined ones and special variables, like `lat: lon: standard_deviation`, `area: mean`, etc.
    def __init__(
        self,
        method: MethodType = None,
        interval: str = None,  # Timedelta/Interval type??
        axis: Union[
            str, Iterable[str]
        ] = None,  #  # axis on which is defined (could be more than one axis??)
        comment: str = None,
        where: str = None,
    ) -> None:
        self.method = method if method is not None else MethodType.UNDEFINED
        self.axis = axis
        self.interval = interval
        self.comment = comment
        self.where = where

    def __eq__(self, other):
        # Comment is not the subject of comparison
        return (
            (self.method is other.method)
            and (self.interval == other.interval)
            and (self.axis == other.axis)
            and (self.where == other.where)
        )

    def __ne__(self, other):
        return not (self == other)

    @classmethod
    def parse(cls, val: str) -> "CellMethod":
        if val is None:
            return None
        interval_start_idx = (
            comment_start_idx
        ) = where_start_idx = par_open_index = par_close_index = np.nan
        if "interval:" in val:
            interval_start_idx = val.find("interval:")
        if "comment:" in val:
            comment_start_idx = val.find("comment:")
        if "where" in val:
            where_start_idx = val.find("where")
        if "(" in val:
            par_open_index = val.find("(")
        if ")" in val:
            par_close_index = val.find(")")

        where_val = interval_val = comment_val = None
        idx_list = [
            interval_start_idx,
            comment_start_idx,
            where_start_idx,
            par_open_index,
            par_close_index,
        ]

        if np.isnan(idx_list).all():
            *axis, method = val.split(": ")
        else:
            *axis, method = (
                item.strip()
                for item in val[: int(np.nanmin(idx_list))].split(": ")
            )

        if not np.isnan(where_start_idx):
            # The case like `time: max where land`
            if not np.isnan(par_open_index):
                where_val = val[where_start_idx + 6 : par_open_index].strip()
            else:
                where_val = val[where_start_idx + 6 :].strip()
        if not np.isnan(interval_start_idx):
            # The case like `time : max (interval: 1hr comment: aaa)`
            # or `time : max (interval: 1hr)`
            interval_ends = int(
                np.nanmin([comment_start_idx, par_close_index])
            )
            if not np.isnan(interval_ends):
                interval_val = val[
                    interval_start_idx + 10 : interval_ends
                ].strip()
            else:
                interval_val = val[interval_start_idx + 10 :].strip()
        if not np.isnan(comment_start_idx):
            comment_val = val[comment_start_idx + 9 : par_close_index].strip()
        return CellMethod(
            method=MethodType(method),
            interval=interval_val,
            axis=axis,
            where=where_val,
            comment=comment_val,
        )

    def __str__(self) -> str:
        res_str = str(self.method)
        if self.axis is not None:
            res_str = ": ".join([*self.axis, res_str])
        if self.where is not None:
            res_str = " ".join([res_str, f"where {self.where}"])
        if (self.interval is not None) and (self.comment is not None):
            res_str = " ".join(
                [
                    res_str,
                    f"(interval: {self.interval} comment: {self.comment})",
                ]
            )
        else:
            if self.interval is not None:
                res_str = " ".join([res_str, f"(interval: {self.interval})"])
            if self.comment is not None:
                res_str = " ".join([res_str, f"({self.comment})"])

        return res_str



import pytest
import xarray as xr

from geokube.backend import open_dataset
from geokube.core.datacube import DataCube


@pytest.fixture
def dataset():
    yield open_dataset(
        "tests/resources/*-single-levels-reanalysis_*",
        pattern=(
            "tests/resources/{dataset}-single-levels-reanalysis_{vars}.nc"
        ),
    )


@pytest.fixture
def dataset_rotated():
    yield open_dataset(
        "tests/resources/rlat-rlon-*",
        pattern="tests/resources/rlat-rlon-{vars}.nc",
    )


@pytest.fixture
def dataset_single_att():
    yield open_dataset(
        "tests/resources/*-single-levels-reanalysis_*",
        pattern="tests/resources/{dataset}-single-levels-reanalysis_{}.nc",
    )


@pytest.fixture
def dataset_idpattern():
    yield open_dataset(
        "tests/resources/era5-single-levels-reanalysis_*",
        pattern=(
            "tests/resources/{dataset}-single-levels-reanalysis_{vars}.nc"
        ),
        id_pattern="std_{units}",
    )


@pytest.fixture
def datacube_with_the_same_standard_name(era5_rotated_netcdf, era5_netcdf):
    era5_rotated_netcdf["TMIN_2M"].attrs["standard_name"] = "std_name1"
    era5_rotated_netcdf["W_SO"].attrs["standard_name"] = "std_name1"
    dc = DataCube.from_xarray(era5_rotated_netcdf)
    yield dc


@pytest.fixture
def rotated_pole_datacube(era5_rotated_netcdf):
    dc = DataCube.from_xarray(era5_rotated_netcdf)
    return dc


@pytest.fixture
def era5_point_domain():
    return xr.open_mfdataset(
        "tests/resources/point_domain*.nc",
        chunks="auto",
        decode_coords="all",
    )


@pytest.fixture
def era5_netcdf():
    yield xr.open_mfdataset(
        "tests/resources/era5-single*.nc", chunks="auto", decode_coords="all"
    )


@pytest.fixture
def era5_globe_netcdf():
    yield xr.open_dataset(
        "tests/resources/globe-era5-single-levels-reanalysis.nc",
        chunks="auto",
        decode_coords="all",
    )


@pytest.fixture
def era5_rotated_netcdf_tmin2m():
    yield xr.open_mfdataset(
        "tests/resources/rlat-rlon-tmin2m.nc",
        chunks="auto",
        decode_coords="all",
    )


@pytest.fixture
def era5_rotated_netcdf_wso():
    yield xr.open_mfdataset(
        "tests/resources/rlat-rlon-wso.nc",
        chunks="auto",
        decode_coords="all",
    )


@pytest.fixture
def era5_rotated_netcdf():
    yield xr.open_mfdataset(
        "tests/resources/rlat-*.nc", chunks="auto", decode_coords="all"
    )


@pytest.fixture
def era5_rotated_netcdf_lat(era5_rotated_netcdf_wso):
    yield era5_rotated_netcdf_wso.lat


@pytest.fixture
def era5_rotated_netcdf_soil(era5_rotated_netcdf_wso):
    yield era5_rotated_netcdf_wso.soil1


@pytest.fixture
def era5_rotated_netcdf_soil_bnds(era5_rotated_netcdf_wso):
    yield era5_rotated_netcdf_wso.soil1_bnds


@pytest.fixture
def nemo_ocean_16():
    dset = xr.open_mfdataset(
        "tests/resources/nemo_ocean_16.nc",
        chunks="auto",
        decode_coords="all",
    )
    # NOTE: there are two time-related coordinates.
    # It is not supported yet to have multiple coordinates of the same AxisType
    dset["vt"].encoding["coordinates"] = "depthv nav_lat nav_lon"
    yield dset.drop(["time_centered", "time_centered_bounds"])



import os
import pytest
from zipfile import ZipFile

from geokube.backend import open_dataset

from tests.fixtures import *
from tests import RES_PATH, RES_DIR, clear_test_res


def test_keeping_files_after_selecting_field(dataset):
    assert "files" in dataset.data
    tp = dataset["tp"]
    assert "files" in tp.data


def test_select_fields_by_name(dataset_idpattern):
    d2m = dataset_idpattern["std_K"]
    assert len(d2m) == 1
    tp = dataset_idpattern["std_m"]
    assert len(tp) == 1


def test_select_fields_by_ncvar(dataset_idpattern):
    d2m = dataset_idpattern["d2m"]
    assert len(d2m) == 1
    tp = dataset_idpattern["tp"]
    assert len(tp) == 1


def test_if_to_dict_produces_json_serializable(dataset, dataset_single_att):
    import json

    _ = json.dumps(dataset.to_dict())
    _ = json.dumps(dataset_single_att.to_dict())

@pytest.mark.skip(
    "Invalidate as in the current version, test don't take into consideration compression and result to be inaccurate"
)
def test_nbytes_estimation(dataset_single_att):
    import os

    clear_test_res()
    d2m = dataset_single_att.sel(
        time={"day": [5, 8], "hour": [1, 2, 3, 12, 13, 14, 22, 23]}
    ).geobbox(north=44, south=39, east=12, west=7)
    precomputed_nbytes = d2m.nbytes
    assert precomputed_nbytes != 0
    os.mkdir(RES_DIR)
    d2m.persist(RES_DIR)
    postcomputed_nbytes = sum(
        [
            os.path.getsize(os.path.join(RES_DIR, f))
            for f in os.listdir(RES_DIR)
        ]
    )
    clear_test_res()
    assert (
        (precomputed_nbytes - postcomputed_nbytes) / postcomputed_nbytes
    ) < 0.25  # TODO: maybe estimation should be more precise


def test_persist_and_return_paths_no_zipping(dataset):
    clear_test_res()
    res_dir = dataset.persist(RES_DIR, zip_if_many=False)
    files = os.listdir(res_dir)
    assert len(files) == 4


def test_persist_and_return_paths_with_zipping(dataset):
    clear_test_res()
    res = dataset.persist(RES_DIR, zip_if_many=True)
    files = os.listdir(RES_DIR)
    assert len(files) == 1
    assert os.path.join(RES_DIR, files[0]) == res
    assert res.endswith(".zip")
    with ZipFile(res, "r") as archive:
        names = archive.namelist()
    for n in names:
        assert RES_DIR not in n
    assert len(names) == 4
    clear_test_res()


def test_persisting_with_empty_one_datacube(dataset_single_att):
    clear_test_res()
    cube0 = dataset_single_att.data.iloc[0].datacube
    cube0 = cube0.sel(time=slice("1-01-01", "1-02-01"))
    dataset_single_att.data.iat[0, 2] = cube0
    path = dataset_single_att.persist(RES_DIR)
    assert ".nc" in path
    assert ".zip" not in path
    clear_test_res()


def test_persisting_with_empty_all_datacube(dataset_single_att):
    clear_test_res()
    cube0 = dataset_single_att.data.iloc[0].datacube
    cube0 = cube0.sel(time=slice("1-01-01", "1-02-01"))
    cube1 = dataset_single_att.data.iloc[1].datacube
    cube1 = cube1.sel(time=slice("1-01-01", "1-02-01"))
    dataset_single_att.data.iat[0, 2] = cube0
    dataset_single_att.data.iat[1, 2] = cube1
    with pytest.warns(
        UserWarning,
        match=r"No files were created while geokube.Dataset persisting!",
    ):
        path = dataset_single_att.persist(RES_DIR)
    assert path is None
    clear_test_res()


def test_attr_str_for_persistance(dataset):
    clear_test_res()
    attr_str = dataset._form_attr_str(dataset.data.iloc[0])
    assert attr_str == "dataset=era5-vars=2_mdt"
    for file in dataset.data.iloc[0]["files"]:
        assert (
            dataset._convert_attributes_to_file_name(
                attr_str, dataset.data.iloc[0]["files"][0]
            )
            == f"dataset=era5-vars=2_mdt-{os.path.basename(file)}"
        )
    clear_test_res()


def test_to_dict_contains_proper_keys(dataset):
    details = dataset.to_dict()
    assert isinstance(details, list)
    assert len(details) == 4
    for d in details:
        assert isinstance(d, dict)
        assert "datacube" in d
        assert "attributes" in d


def test_to_dict_contain_attributes(dataset):
    details = dataset.to_dict()
    for d in details:
        assert "attributes" in d
        assert len(d["attributes"]) == 2
        assert d["attributes"]["vars"] in {"total_precipitation", "2_mdt"}
        assert d["attributes"]["dataset"] in {"other-era5", "era5"}


def test_to_dict_contains_proper_datacube_fields_rot(dataset_rotated):
    details = dataset_rotated.to_dict()
    d = details[0]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "air_temperature" in fields.keys()

    d = details[1]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "lwe_thickness_of_moisture_content_of_soil_layer" in fields.keys()


def test_to_dict_contains_proper_datacube_fields(dataset):
    details = dataset.to_dict()
    d = details[0]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "d2m" in fields.keys()

    d = details[1]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "tp" in fields.keys()

    d = details[2]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "d2m" in fields.keys()

    d = details[3]
    fields = d["datacube"]["fields"]
    assert isinstance(fields, dict)
    assert len(fields) == 1
    assert "tp" in fields.keys()



from __future__ import annotations

__all__ = (
    "open_datacube",
    "open_dataset",
)

import glob
import logging
import os
import pickle
from string import Formatter
from typing import Any, Hashable, Mapping, Optional

import dask
import pandas as pd
import xarray as xr
from intake.source.utils import reverse_format
import rioxarray

import geokube.backend.base
import geokube.core.datacube
import geokube.core.dataset
from geokube.utils.hcube_logger import HCubeLogger

LOG = HCubeLogger(name="netcdf.py")

FILES_COL = geokube.core.dataset.Dataset.FILES_COL
DATACUBE_COL = geokube.core.dataset.Dataset.DATACUBE_COL


def _get_engine(path: list | str):
    if isinstance(path, list):
        if len(path) > 0:
            path = path[0]
        else:
            raise ValueError("empty path list provided!")
    if isinstance(path, str):
        _, ext = os.path.splitext(path)
    else:
        raise TypeError(f"unsupported path type: `{type(path)}`")
    if ext == ".tif":
        return "rasterio"
    elif ext == ".nc":
        return "netcdf4"
    elif ext == ".jp2":
        return "rasterio"
    elif ext == ".zarr":
        return "zarr"
    elif path.startswith('http') or path.startswith('https'):
        return "zarr"
    else:
        raise ValueError(
            f"there is not engine associated with the extension `{ext}`"
        )


def _read_cache(cache_path: str):
    cached = None
    if os.path.exists(cache_path):
        with open(cache_path, "rb") as f:
            cached = pickle.load(f)
    return cached


def _write_cache(ds, cache_path: str):
    with open(cache_path, "wb") as f:
        pickle.dump(ds, f)


def open_datacube(
    path: str,
    id_pattern: Optional[str] = None,
    mapping: Optional[Mapping[str, Mapping[str, str]]] = None,
    metadata_caching: bool = False,
    metadata_cache_path: str = None,
    **kwargs,  # optional kw args for xr.open_mfdataset
) -> geokube.core.datacube.DataCube:
    # TODO: incremental metadata caching
    # we could load the cache file and compare the files from the paths with the files
    # in the dataframe cached and read only the files that are not in the cache
    if metadata_caching:
        ds = _read_cache(metadata_cache_path)
        if ds is not None:
            return ds
    engine = kwargs.pop('engine', None) or _get_engine(path)
    if engine == "netcdf4":
        kwargs.setdefault("decode_coords", "all")
    if engine == "zarr":
        kwargs.setdefault("decode_coords", "all")
    ds = geokube.core.datacube.DataCube.from_xarray(
        xr.open_mfdataset(path, engine=engine, **kwargs),
        id_pattern=id_pattern,
        mapping=mapping,
    )
    if metadata_caching:
        _write_cache(ds, metadata_cache_path)

    return ds


def _get_ds_attrs_names(pattern):
    fmt = Formatter()
    # get the dataset attrs from the pattern
    ds_attr_names = [i[1] for i in fmt.parse(pattern) if i[1]]
    return ds_attr_names


def _get_df_from_files_list(files, pattern, ds_attr_names):
    l = []
    for f in files:
        d = reverse_format(pattern, f)
        d[FILES_COL] = f
        l.append(d)
    df = pd.DataFrame(l)
    if len(l) == 0:
        raise ValueError(f"No files found for the provided path!")
    # unique index for each dataset attribute combos - we create a list of files
    df = df.groupby(ds_attr_names)[FILES_COL].apply(list).reset_index()
    df = df.set_index(ds_attr_names)
    return df


def open_dataset(
    path: str,
    pattern: str,
    id_pattern: Optional[str] = None,
    mapping: Optional[Mapping[str, Mapping[str, str]]] = None,
    metadata_caching: bool = False,
    metadata_cache_path: str = None,
    ds_attr_mapping: Mapping[
        Hashable, Any
    ] = None,  # dataset attributes mapping - TBA
    # { 'attr_name1': { 'id': ..., 'description': ... }, ...}
    ncvars_mapping: Mapping[
        Hashable, Any
    ] = None,  # netcdf variables mapping - TBA
    # { 'ncvar_name1': {'id': ..., 'description': }, ...}
    delay_read_cubes: bool = False,  # when True the method will not create datacubes when opening a dataset; this is useful
    # when the number of rows is really high and the number of files per row is low
    # (e.g CMIP, CORDEX, observations). The datacube will be read when trying to accessing it
    load_files_on_persistance: bool = True,
    **kwargs,  # optional kw args for xr.open_mfdataset
) -> geokube.core.dataset.Dataset:
    # incremental metadata caching:
    # load the cache file and compare the files from the paths with the files
    # in the dataframe cached and read only the files that are not in the cache
    ds_attr_names = _get_ds_attrs_names(pattern)
    if metadata_caching:
        if metadata_cache_path is None:
            raise ValueError(
                "If `metadata_caching` set to True, `metadata_cache_path`"
                " argument needs to be provided!"
            )
        cached_ds = _read_cache(metadata_cache_path)
        if cached_ds is not None:
            # cached_files = list(cached_ds.reset_index()[FILES_COL])
            # cached_files = [
            #     item for sublist in cached_files for item in sublist
            # ]
            # TODO: below glob takes too much time
            # while cache loading, e.g. for gutta-visir
            # files = glob.glob(path)  # all files
            # not_cached_files = list(set(files) - set(cached_files))
            # if len(not_cached_files) == 0:  # there are no new files
            return geokube.core.dataset.Dataset(
                hcubes=cached_ds.reset_index(),
                load_files_on_persistance=load_files_on_persistance,
            )

            # there are new files we need to update the cache
            # we consider the case we only add files
            # we should consider also when files are deleted (e.g. rolling archives)
            not_cached_ds = _get_df_from_files_list(
                not_cached_files, pattern, ds_attr_names
            )
            for i in not_cached_ds.index:
                # check if index i is in the cached
                # if index exists update the datacube  (merge __FILES column and open_datacube)
                # if index does not exist add a new row
                if i in cached_ds.index:
                    new_files = [
                        *cached_ds[FILES_COL][i],
                        *not_cached_ds[FILES_COL][i],
                    ]
                    if not load_files_on_persistance:
                        cube = None
                    elif delay_read_cubes:
                        cube = dask.delayed(open_datacube)(
                            path=new_files,
                            id_pattern=id_pattern,
                            mapping=mapping,
                            **kwargs,
                        )
                    else:
                        cube = open_datacube(
                            path=new_files,
                            id_pattern=id_pattern,
                            mapping=mapping,
                            **kwargs,
                        )
                    cached_ds.loc[i] = {
                        FILES_COL: new_files,
                        DATACUBE_COL: cube,
                    }
                elif i in not_cached_ds.index:
                    not_cached_files = not_cached_ds[FILES_COL][i]
                    if delay_read_cubes:
                        cube = dask.delayed(open_datacube)(
                            path=not_cached_files,
                            id_pattern=id_pattern,
                            mapping=mapping,
                            **kwargs,
                        )
                    else:
                        cube = open_datacube(
                            path=not_cached_files,
                            id_pattern=id_pattern,
                            mapping=mapping,
                            **kwargs,
                        )
                    # To find better way
                    cached_ds.loc[i, :] = [not_cached_files, None]
                    cached_ds.loc[i, :] = [not_cached_files, cube]
                else:  # the row from cached_ds should be deleted!
                    pass

            _write_cache(cached_ds, metadata_cache_path)
            return geokube.core.dataset.Dataset(
                hcubes=cached_ds.reset_index(),
                load_files_on_persistance=load_files_on_persistance,
            )

    # if cache is not True or cache file is not available proceed with reading files in paths
    files = glob.glob(path)  # all files
    df = _get_df_from_files_list(files, pattern, ds_attr_names)
    cubes = []
    if not load_files_on_persistance:
        pass
    elif delay_read_cubes:
        for i in df.index:
            cubes.append(
                dask.delayed(open_datacube)(
                    path=df[FILES_COL][i],
                    id_pattern=id_pattern,
                    mapping=mapping,
                    **kwargs,
                )
            )
    else:
        for i in df.index:
            cubes.append(
                open_datacube(
                    path=df[FILES_COL][i],
                    id_pattern=id_pattern,
                    mapping=mapping,
                    **kwargs,
                )
            )  # we do not need to enable caching here!
    if load_files_on_persistance:
        df[DATACUBE_COL] = cubes
    else:
        df[DATACUBE_COL] = None
    # write cache if cache file does not exist and caching is true
    if metadata_caching:
        _write_cache(df, metadata_cache_path)

    return geokube.core.dataset.Dataset(
        hcubes=df.reset_index(),
        load_files_on_persistance=load_files_on_persistance,
    )



import os
import json
import logging
import re
import tempfile
import uuid
import warnings
from collections import defaultdict
from enum import Enum
from html import escape
from itertools import chain
from numbers import Number
from string import Template
from types import MethodType
from typing import (
    Any,
    Callable,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import xarray as xr
import math

from ..utils.decorators import geokube_logging
from ..utils.hcube_logger import HCubeLogger
from ..utils import util_methods
from .errs import EmptyDataError
from .axis import Axis, AxisType
from .coord_system import GeogCS, RegularLatLon
from .domain import Domain, DomainType
from .enums import RegridMethod
from .field import Field
from .domainmixin import DomainMixin

IndexerType = Union[slice, List[slice], Number, List[Number]]


# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring


# TODO: Not a priority
# dc = datacube.set_id_pattern('{standard_name}')
# dc['air_temperature']
# dc['latitude']
#
class DataCube(DomainMixin):
    __slots__ = (
        "_fields",
        "_domain",
        "_properties",
        "_encoding",
        "_ncvar_to_name",
    )

    _LOG = HCubeLogger(name="DataCube")

    def __init__(
        self,
        fields: List[Field],
        properties: Mapping[Any, Any],
        encoding: Mapping[Any, Any],
    ) -> None:
        if len(fields) == 0:
            warnings.warn("No fields provided for the DataCube!")
            self._fields = {}
            self._domain = None
            self._ncvar_to_name = None
        else:
            self._ncvar_to_name = {f.ncvar: f.name for f in fields}
            self._fields = {f.name: f for f in fields}
            self._domain = Domain.merge([f.domain for f in fields])
        self._properties = properties if properties is not None else {}
        self._encoding = encoding if encoding is not None else {}

    @property
    def properties(self) -> dict:
        return self._properties

    @property
    def encoding(self) -> dict:
        return self._encoding

    @property
    def domain(self) -> Domain:
        return self._domain

    @property
    def fields(self) -> dict:
        return self._fields

    @property
    def nbytes(self) -> int:
        return sum(field.nbytes for field in self.fields.values())

    def __len__(self):
        return len(self._fields)

    def __contains__(self, key: str) -> bool:
        return (
            (key in self._fields)
            or (key in self._ncvar_to_name)
            or (key in self._domain)
        )

    @geokube_logging
    def __getitem__(
        self,
        key: Union[
            Iterable[str], Iterable[Tuple[str, str]], str, Tuple[str, str]
        ],
    ):
        if isinstance(key, str) and (
            (key in self._fields) or key in self._ncvar_to_name
        ):
            return self._fields.get(
                key, self._fields.get(self._ncvar_to_name.get(key))
            )
        elif isinstance(key, Iterable) and not isinstance(key, str):
            return DataCube(
                fields=[self[k] for k in key],
                properties=self.properties,
                encoding=self.encoding,
            )
        else:
            item = self.domain[key]
            if item is None:
                raise KeyError(
                    f"Key `{key}` of type `{type(key)}` is not found in the"
                    " DataCube"
                )
            return item

    def __next__(self):
        for f in self._fields.values():
            yield f
        raise StopIteration

    def __repr__(self) -> str:
        return self.to_xarray(encoding=False).__repr__()

    #        return formatting.array_repr(self.to_xarray())

    def _repr_html_(self):
        return self.to_xarray(encoding=False)._repr_html_()
        # if OPTIONS["display_style"] == "text":
        #     return f"<pre>{escape(repr(self.to_xarray()))}</pre>"
        # return formatting_html.array_repr(self)

    @geokube_logging
    def geobbox(
        self,
        north=None,
        south=None,
        west=None,
        east=None,
        top=None,
        bottom=None,
    ):
        return DataCube(
            fields=[
                self._fields[k].geobbox(
                    north=north,
                    south=south,
                    east=east,
                    west=west,
                    top=top,
                    bottom=bottom,
                )
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def locations(
        self,
        latitude,
        longitude,
        vertical: Optional[List[Number]] = None,
    ):
        return DataCube(
            fields=[
                self._fields[k].locations(
                    latitude=latitude, longitude=longitude, vertical=vertical
                )
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def sel(
        self,
        indexers: Mapping[Union[Axis, str], Any] = None,
        method: str = None,
        tolerance: Number = None,
        drop: bool = False,
        roll_if_needed: bool = True,
        **indexers_kwargs: Any,
    ) -> "DataCube":  # this can be only independent variables
        fields = []
        for k in self._fields.keys():
            try:
                fields.append(
                    self._fields[k].sel(
                        indexers=indexers,
                        roll_if_needed=roll_if_needed,
                        method=method,
                        tolerance=tolerance,
                        drop=drop,
                        **indexers_kwargs,
                    )
                )
            except EmptyDataError as err:
                DataCube._LOG.info(f"skipping field `{k}` due to `{err}`")
                continue
        return DataCube(
            fields=fields,
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def interpolate(
        self, domain: Domain, method: str = "nearest"
    ) -> "DataCube":
        return DataCube(
            fields=[
                self._fields[k].interpolate(domain=domain, method=method)
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def resample(
        self,
        operator: Union[Callable, MethodType, str],
        frequency: str,
        **resample_kwargs,
    ) -> "DataCube":
        return DataCube(
            fields=[
                self._fields[k].resample(
                    operator=operator, frequency=frequency, **resample_kwargs
                )
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def average(self, dim: str | None = None) -> "DataCube":
        return DataCube(
            fields=[
                self._fields[k].average(dim=dim) for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def to_regular(
        self,
    ) -> "DataCube":
        return DataCube(
            fields=[self._fields[k].to_regular() for k in self._fields.keys()],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def extract_polygons(self, geometry, crop=True, return_mask=False):
        return DataCube(
            fields=[
                self._fields[k].extract_polygons(
                    geometry=geometry, crop=crop, return_mask=return_mask
                )
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    @geokube_logging
    def regrid(
        self,
        target: Union[Domain, "Field"],
        method: Union[str, RegridMethod] = "bilinear",
        weights_path: Optional[str] = None,
        reuse_weights: bool = True,
    ) -> "DataCube":
        return DataCube(
            fields=[
                self._fields[k].regrid(
                    target=target,
                    method=method,
                    weights_path=weights_path,
                    reuse_weights=reuse_weights,
                )
                for k in self._fields.keys()
            ],
            properties=self.properties,
            encoding=self.encoding,
        )

    def to_geojson(self, target=None):
        if self.domain.type is DomainType.POINTS:
            if self.latitude.size != 1 or self.longitude.size != 1:
                raise NotImplementedError(
                    "'self.domain' must have exactly 1 point"
                )
            units = {
                field.name: str(field.units) for field in self.fields.values()
            }
            coords = [self.longitude.item(), self.latitude.item()]
            result = {"type": "FeatureCollection", "features": []}
            for time in self.time.values.flat:
                time_ = pd.to_datetime(time).strftime("%Y-%m-%dT%H:%M")
                feature = {
                    "type": "Feature",
                    "units": units,
                    "geometry": {"type": "Point", "coordinates": coords},
                    "properties": {"time": time_},
                }
                for field in self.fields.values():
                    field.load()
                    try:
                        value = (
                            field.sel(time=time_)
                            if field.time.size > 1
                            else field
                        )
                    except EmptyDataError:
                        continue
                    else:
                        feature["properties"][field.name] = float(value)
                result["features"].append(feature)
        elif (
            self.domain.type is DomainType.GRIDDED or self.domain.type is None
        ):
            # HACK: The case `self.domain.type is None` is included to be able
            # to handle undefined domain types temporarily.
            if self.time.size == 1:
                result = {}
            else:
                raise NotImplementedError(
                    f"multiple times are not supported for geojson"
                )
            cube = (
                self
                if isinstance(self.domain.crs, GeogCS)
                else self.to_regular()
            )
            for field in self.fields.values():
                field.load()
            axis_names = cube.domain._axis_to_name
            units = {
                field.name: str(field.units) for field in self.fields.values()
            }
            lon_min = self.longitude.min().item()
            lat_min = self.latitude.min().item()
            lon_max = self.longitude.max().item()
            lat_max = self.latitude.max().item()      
            grid_x, grid_y = cube.domain._infer_resolution()
            grid_x = grid_x/2.0
            grid_y = grid_y/2.0
            for time in self.time.values.flat:
                time_ = pd.to_datetime(time).strftime("%Y-%m-%dT%H:%M")
                time_data = {
                    "type": "FeatureCollection",
                    "date": time_,
                    "bbox": [
                        lon_min,  # West
                        lat_min,  # South
                        lon_max,  # East
                        lat_max,  # North
                    ],
                    "units": units,
                    "features": [],
                }
                for lat in cube.latitude.values.flat:
                    for lon in cube.longitude.values.flat:
                        idx = {
                            axis_names[AxisType.LATITUDE]: lat,
                            axis_names[AxisType.LONGITUDE]: lon,
                        }
                        # if self.time.shape:
                        if self.time.size > 1:
                            idx[axis_names[AxisType.TIME]] = time_
                        # TODO: Check whether this works now:
                        # this gives an error if only 1 time is selected before to_geojson()
                        # Polygon:
                        # for each lat/lon we have to define a polygon with lan/lon centered
                        # the cell length depends on the grid resolution (that should be computed)
                        lonv = lon.item()
                        latv = lat.item()
                        lon_lower = np.clip(lonv - grid_x, a_min=lon_min, a_max=lon_max)
                        lat_upper = np.clip(latv + grid_y, a_min=lat_min, a_max=lat_max)
                        lon_upper = np.clip(lonv + grid_x, a_min=lon_min, a_max=lon_max)
                        lat_lower = np.clip(latv - grid_y, a_min=lat_min, a_max=lat_max)
                        feature = {
                            "type": "Feature",
                            "geometry": {
                                "type": "Polygon",
                                "coordinates":[  
                                   [ [lon_lower, lat_upper], [lon_upper, lat_upper],
                                     [lon_upper, lat_lower], [lon_lower, lat_lower],
                                     [lon_lower, lat_upper]
                                   ]
                                ]
                            },
                            "properties": {},
                        }
                        for field in cube.fields.values():
                            try:
                                value = field.sel(indexers=idx)
                                value_ = float(value)
                                if math.isnan(value):
                                    value_ = None                                
                            except EmptyDataError:
                                continue
                            except ValueError:
                                try:
                                    value_ = value.item()
                                except AttributeError:
                                    value_ = value                                
                            else:
                                feature["properties"][field.name] = value_                     
                        time_data["features"].append(feature)
                result = time_data
        else:
            raise NotImplementedError(
                f"'self.domain.type' is {self.domain.type}, which is currently"
                " not supported"
            )

        if target is not None:
            with open(target, mode="w") as file:
                json.dump(result, file, indent=4)

        return result

    def to_image(
        self, 
        filepath, 
        width, 
        height, 
        dpi=100, 
        format='png', 
        transparent=True, 
        bgcolor='FFFFFF',
        cmap='RdBu_r',
        projection=None,
        vmin=None,
        vmax=None
    ):
        # TODO: support multiple fields
        if len(self.fields) > 1:
            raise ValueError("to_image support only 1 field")
        else:
            next(iter(self.fields.values())).to_image(filepath, width, height, dpi, format, transparent, bgcolor, cmap, projection, vmin, vmax)

    @classmethod
    @geokube_logging
    def from_xarray(
        cls,
        ds: xr.Dataset,
        id_pattern: Optional[str] = None,
        mapping: Optional[Mapping[str, str]] = None,
    ) -> "DataCube":
        fields = []
        #
        # we assume that data_vars contains only variable + ancillary
        # and coords all coordinates, grid_mapping and so on ...
        # TODO ancillary variables
        #
        for dv in ds.data_vars:
            fields.append(
                Field.from_xarray(
                    ds, ncvar=dv, id_pattern=id_pattern, mapping=mapping
                )
            )
        # Issue https://github.com/opengeokube/geokube/issues/221
        attrs = ds.attrs.copy()
        encoding = ds.encoding.copy()
        attrs.pop("_NCProperties", None)
        encoding.pop("_NCProperties", None)
        return DataCube(fields=fields, properties=attrs, encoding=encoding)

    @geokube_logging
    def to_xarray(self, encoding=True):
        xarray_fields = [
            f.to_xarray(encoding=encoding) for f in self.fields.values()
        ]
        dset = xr.merge(
            xarray_fields, join="outer", combine_attrs="no_conflicts", compat='override',
        )
        dset.attrs = self.properties
        dset.encoding = self.encoding
        return dset

    @geokube_logging
    def to_netcdf(self, path, encoding: bool = True):
        self.to_xarray(encoding=encoding).to_netcdf(path=path)

    def to_zarr(self, path, encoding: bool = True, **kwargs):
        kube = self.to_xarray(encoding=encoding)
        for var in kube:
            if 'chunks' in kube[var].encoding.keys():
                del kube[var].encoding['chunks']
        kube.chunk('auto').to_zarr(path,**kwargs)

    @geokube_logging
    def to_csv(self, path, encoding: bool = True):
        self.to_xarray(encoding=encoding).to_dataframe().drop(columns=['crs_latitude_longitude']).to_csv(path)

    def persist(self, path=None) -> str:
        if path is None:
            path = os.path.join(
                tempfile.gettempdir(), f"{str(uuid.uuid4())}.nc"
            )
        if os.path.isdir(path):
            path = os.path.join(path, f"{str(uuid.uuid4())}.nc")
        if not path.endswith(".nc"):
            self._LOG.warn(
                f"Provided persistance path: `{path}` has not `.nc` extension."
                " Adding automatically!"
            )
            warnings.warn(
                f"Provided persistance path: `{path}` has not `.nc` extension."
                " Adding automatically!"
            )
            path = path + ".nc"
        self.assert_not_empty()
        self.to_netcdf(path)
        return path

    def to_dict(self, unique_values=False) -> dict:
        return {
            "domain": self.domain.to_dict(unique_values),
            "fields": {k: v.to_dict() for k, v in self.fields.items()},
        }

    def assert_not_empty(self):
        if not len(self):
            self._LOG.warn("No fields in DataCube")
            raise EmptyDataError("No fields in DataCube!")
        for fname, field in self.fields.items():
            if 0 in field.shape:
                self._LOG.warn(
                    f"One of coordinate is empty for the field `{fname}`."
                    f" Shape=`{field.shape}`. Dimensions=`{field.dim_names}`!"
                )
                raise EmptyDataError(
                    f"One of coordinate is empty for the field `{fname}`."
                    f" Shape=`{field.shape}`. Dimensions=`{field.dim_names}`!"
                )
